# An application of our stable module methods to the running example of the 
# main text (Eq. 1).

import numpy as np
import ODEToDigraph as od

# Define the following ordering of variables
u = 0
w = 1
x = 2
y = 3
z = 4
f = 5

numVars = 6

# List of variable names coinciding with the above ordering
keys = ["u","w","x","y","z","f"]

# Define initial conditions; these do not affect stable module results
u0 = [0]*numVars
u0[u] = 0.5
u0[w] = 0.5
u0[x] = 0.5
u0[y] = 0.5
u0[z] = 0.5
u0[f] = 0.0

def update( uu, t ):
    d = [0]*numVars 
    d[u] = 1/(1+uu[z]) - uu[u]**3
    d[w] = uu[y] - uu[w] * 0.5
    d[x] = (uu[w]*uu[z] + uu[w]+0.25) / (uu[w]*uu[z]+uu[w]*0.5+uu[z]*0.5+0.25) - uu[x]
    d[y] = uu[x]/(uu[x]+0.5) - uu[y]
    d[z] = uu[x]*uu[f] - uu[z]
    d[f] = 0 # unknown evolution; will not use, so no need to calculate a change
    
    return d 

# The variables u, x, and y have bounded regulatory functions, and will not
# exceed 1/3, 2, and 1, respectively (asymptotically). Similarly, w will not 
# exceed y asymptotically. We choose 3 as an upper bound for these variables.
# The evolution of z, however, is not known, because the function f is not
# known, so we choose a much larger value for the upper bound of z. Ideally,
# this should be infinite, but 100000 is sufficiently large that the difference
# in the regulatory functions dependent on z is negligible.
maxVals = np.ones( numVars )*3
maxVals[z] = 100000
minVals = np.zeros( numVars )

tol = [0.001]*numVars

thresh, high, stable = od.stableModuleSearch( update, u0, minVals, maxVals, tol )
print()
print()
print("==========================================")
print()
for t,s,h in zip( thresh, stable, high ):
    print("-------------------------")
    od.printModuleNodes( t, h, keys = keys )
    print("-------------------------")
