import networkx as nx
import numpy as np
import matplotlib.pyplot as plt

import scipy as sp
from scipy.integrate import odeint
import scipy.optimize
from itertools import combinations

'''
Create the signed regulatory network of the 1st order ode update at time t0

----------
- Inputs -
----------
update: 1st order ode regulatory functions (takes and returns lists)
        e.g., dx/dt = update(x,t), where x is a vector
a:      lower testing bounds (list)
b:      upper testing bounds (list)
c:      value of nodes not being explicitly tested (list)
t0:     time at which to evaluate the regulatory network (unimportant for 
        autonomous systems)

-----------
- Outputs -
-----------
G: A networkx digraph representing the regulatory network; edge weights are
   either -1 or 1, and nodes are labeled by input index

---------
- Notes -
---------
We assume that, on the interval [a[i],b[i]], update is strictly monotonic in 
x[i] or independent of x[i]. We compare update(a[i]) to update(b[i]) to build a
signed digraph, evaluating all other points at c. If c is None, we take c[i]
to be the midpoint of [a[i],b[i]]
'''
def ODEToDigraph( update, a, b, c = None, t0 = 0 ):
    G = nx.DiGraph()
    
    if c is None:
        c = [( xb + xa ) / 2.0 for xb, xa in zip( a, b )]
    for i in range( len( a ) ):
        # Find effects of x[i]
        xmax = c.copy()
        xmin = c.copy()
        xmax[i] = b[i]
        xmin[i] = a[i]
        
        uxmax = update( xmax, t0 ) 
        uxmin = update( xmin, t0 )
        
        for j in range( len( uxmax ) ):
            # Find effect of x[i] on x[j]
            if uxmax[j] > uxmin[j]:
                G.add_edge( i, j, weight = 1 )
            elif uxmax[j] < uxmin[j]:
                G.add_edge( i, j, weight = -1 )
            # else they are equal or None; do not add an edge
    return G


'''
Find graph cycles and their signs. Used to identify positive cycles in the
regulatory network.

----------
- Inputs -
----------
G: networkx digraph with edge weights (weights equal to +/-1)

-----------
- Outputs -
-----------
cs: a list of elements the form (cycleSign, c), where cycleSign is the product
    of edge weights in the cycle c; cs includes all simple cycles of G
'''
def cycleSigns( G ):
    cyc = nx.simple_cycles( G )
    cs = []
    for c in cyc:
        cycleSign = 1
        n_prev = c[-1]
        for n in c:
            cycleSign *= G[n_prev][n]['weight']
            n_prev = n
        cs.append( [cycleSign, c] )
    
    return cs

'''
Generate two "worst case" subsystems given an edge list of a subgraph in the 
regulatory network.

----------
- Inputs -
----------
update:            1st order ode regulatory functions (takes and returns lists)
G:                 regulatory network of update (see ODEToDigraph)
subSystemEdgeList: list of edges that form a subgraph for which we will find the
                   two "worst case" subsystems; assume all nodes reachable from
                   first node of first edge
xmin, xmax:        lists of physical min/max values of variables

-----------
- Outputs -
-----------
dxModA:     regulatory function of system A
dxModB:     regulatory function of system B
highNodesA: list of "high" nodes for system A
highNodesB: list of "high" nodes for system B

---------
- Notes -
---------
dxmodA/B take vectors of the same length as the input to update, and
return vectors of the same length as update. Nodes excluded from these
modified systems have np.nan values returned.
'''
def worstCaseSystem( update, G, subSystemEdgeList, xmin, xmax ):
    highNodesA = set() # = lowNodesB
    highNodesB = set() # = lowNodesA
    ssNodes = set()
    startFlag = True
    
    SSG = nx.DiGraph( subSystemEdgeList )
    walkEdges = list( nx.edge_dfs( SSG ) )
    goodEdges = walkEdges.copy()

    for e in walkEdges:
        u = e[0]
        v = e[1]
        s = G[u][v]['weight']
        
        if not (u,u) in goodEdges:
            goodEdges.append( (u,u) )
        if not (v,v) in goodEdges:
            goodEdges.append( (v,v) )        
            
        # This condition should only be triggered once because u(i+1) = vi!
        if startFlag:
            highNodesA.add( u )
            startFlag = False
        
        ssNodes.add( u )
        uHA = u in highNodesA
        uHB = u in highNodesB
        
        if uHA:
            if s == 1:
                highNodesA.add( v )
            elif s == -1:
                highNodesB.add( v )
        elif uHB:
            if s == 1:
                highNodesB.add( v )
            elif s == -1:
                highNodesA.add( v )
            
    nodeNumber = len( ssNodes )
    Nx = len( xmin )
    xA = np.array( [[np.nan]*Nx] * Nx )
    xB = np.array( [[np.nan]*Nx] * Nx )
    
    # xA[i,j] is the worst case value for the j -> i edge in system A
    # xB is similar
    
    # Now, we have two sets of high nodes
    for n in ssNodes:
        for u, v, data in G.in_edges( n, data = True ):
            # v will always be n
            
            # we don't worry about edges handled elsewhere
            if (u, v) in subSystemEdgeList or u == v:                
                continue
            
            # If n is high in A and u -| n, then worst case A is max u (1)
            # If n is high in A and u -> n, then worst case A is min u (2)
            # If n is low  in A and u -| n, then worst case A is min u (3)
            # If n is low  in A and u -> n, then worst case A is max u (4)
            
            # If n is high in A and u -| n, then worst case B is min u (5)
            # If n is high in A and u -> n, then worst case B is max u (6)
            # If n is low  in A and u -| n, then worst case B is max u (7)
            # If n is low  in A and u -> n, then worst case B is min u (8)
            
            if n in highNodesA:
                if data['weight'] == 1: # 2, 6
                    xA[n,u] = xmin[u]
                    xB[n,u] = xmax[u]
                    if u in highNodesA:
                        goodEdges.append( (u,v) )
                elif data['weight'] == -1: # 1, 5
                    xA[n,u] = xmax[u]
                    xB[n,u] = xmin[u]
                    if u in highNodesB:
                        goodEdges.append( (u,v) ) 
                else:
                    print("Unrecognized Edge Weight!")
            elif n in highNodesB:
                if data['weight'] == 1: # 4, 8
                    xB[n,u] = xmin[u]
                    xA[n,u] = xmax[u]
                    if u in highNodesB:
                        goodEdges.append( (u,v) )
                elif data['weight'] == -1: # 3, 7
                    xB[n,u] = xmax[u]
                    xA[n,u] = xmin[u]
                    if u in highNodesA:
                        goodEdges.append( (u,v) )
                else:
                    print("Unrecognized Edge Weight!")
            
            
    # Now that we have determined worst case values, we use these to define
    # the worst case subsystems
    def dxModA( x, t ):
        dx = [np.nan] * len( x )
        for onode in range( len( x ) ):
            if onode in ssNodes:
                xmod = list( xA[onode,:] )
                for inode in range( len( x ) ):
                    if inode == onode or (inode,onode) in subSystemEdgeList:
                    #if (inode,onode) in goodEdges:
                        xmod[inode] = x[inode]
                dx[onode] = update( xmod, t )[onode]
        return dx
    
    def dxModB( x, t ):
        dx = [np.nan] * len( x )
        for onode in range( len( x ) ):
            if onode in ssNodes:
                xmod = list( xB[onode,:] )
                for inode in range( len( x ) ):
                    if inode == onode or (inode,onode) in subSystemEdgeList:
                    #if (inode,onode) in goodEdges:
                        xmod[inode] = x[inode]
                dx[onode] = update( xmod, t )[onode]
        return dx
    
    return dxModA, dxModB, highNodesA, highNodesB

'''
Find the characteristic of a system at a given value of an input node and a 
topological sort of its regulatory network.

----------
- Inputs -
----------
wcs:       1st order (worst case) ode regulatory functions (takes and returns 
           lists)
nodeOrder: a topological sort of the variables in wcs; includes the input and 
           output nodes as nodeOrder[0] and nodeOrder[-1], respectively
xin:       the value of the input node (nodeOrder[0])
x0:        initial values of other nodes (should not influence result)
t0:        time at which to evaluate the characteristic (not needed for 
           autonomous systems)

-----------
- Outputs -
-----------
x: the value of the characteristic of wcs evaluated at nodeOrder[0] = xin
'''
def characteristic( wcs, nodeOrder, xin, x0, t0 ):
    
    x = x0.copy()
    
    firstNode = True
    for node in nodeOrder:
        if firstNode:
            x[node] = xin
            firstNode = False
        else:
            def dx( xout ):
                x[node] = xout
                return wcs( x, t0 )[node]
            x[node] = sp.optimize.root( dx, x[node] ).x[0]
    
    return x

'''
Find steady states for wcs using MIOS and optionally plot the system 
characteristic.

----------
- Inputs -
----------
wcs:       1st order (worst case) ode regulatory functions (takes and returns 
           lists)
G:         the networkx regulatory digraph of wcs
nodeList: an initial ordering of the variables in wcs (should not matter)
xin:       the value of the input node (nodeOrder[0])
x0:        initial values of nodes (should not influence result)
t0:        time at which to evaluate the characteristic (not needed for 
           autonomous systems)

xmin,xmax,tol:      lists that determine the search space for each node, should
                    it be chosen as the MIOS input node.
plotCharacteristic: determines whether a characteristic plot is shown
'''
def MIOSThreshAlgebraic( wcs, G, nodeList, x0, xmin, xmax, tol, t0 = 0, plotCharacteristic = False ):
    
    # The starting node must leave wcs acyclic if removed. We will try each node
    # in nodeOrder sequentially. Once a good starting node is found, we 
    # topologically sort the rest and append an output node to obtain the new 
    # nodeOrder.
    goodStartNode = False
    startNodeAttempt = 0
    while not goodStartNode and startNodeAttempt < len( nodeList ):
        K = G.copy()
        try:
            K.remove_node( nodeList[0] )
        except Exception:
            pass
        K.remove_edges_from( K.selfloop_edges() )
        if nx.is_directed_acyclic_graph( K ):
            goodStartNode = True
            nodeOrder = [nodeList[0]] + nx.topological_sort( K ) + [nodeList[0]]
        else:
            nodeList = [nodeList[i-1] for i in range( len( nodeList ) )]
            startNodeAttempt += 1
    if not goodStartNode:
        #print("Cannot find a good start node! Topological MIOS is impossible.")
        return [],[]
    
    roots = []
    slopeSigns = []
    allSigns = []
    
    # The search space for characteristic roots
    xx = np.linspace( xmin[nodeOrder[0]], 
                      xmax[nodeOrder[0]], 
                      ( xmax[nodeOrder[0]] - xmin[nodeOrder[0]] ) 
                        / tol[nodeOrder[0]] + 1 )

    lastSign = None

    # We find characteristic roots by a simple brute force approach.
    # To maintain consistency between different variables, we don't interpolate.
    for x in xx:
        x_sol = characteristic( wcs, nodeOrder, x, x0, t0 )
        
        for i in range( len( x_sol ) ):
            if not i in nodeOrder:
                x_sol[i] = np.nan
        
        xout = x_sol[nodeOrder[0]]
        
        sign = xout - x
        
        if plotCharacteristic: 
            allSigns.append(sign)
        if lastSign is not None:
            if lastSign > 0 and sign <= 0:
                roots.append( x_sol )
                slopeSigns.append( -1 )
            elif lastSign < 0 and sign >= 0:
                roots.append( x_sol )
                slopeSigns.append( 1 )
        lastSign = sign
    if plotCharacteristic:
        if min( allSigns ) == 0 and max( allSigns ) == 0 or all( np.isnan( a ) for a in allSigns ):
            print( "Ill-defined system. Skipping Plot." )
        elif roots != []:
            plt.plot( xx, allSigns, '-+' )
            plt.show()
    
    return roots, slopeSigns

'''
Filter out MIOS roots with variable values outside a given range

----------
- Inputs -
----------
roots: list of roots (each root is a list of values) to be filtered
signs: list of signs to be filtered (according to roots)
minVals,maxVals: list of bounds for each variable

-----------
- Outputs -
-----------
roots_cut: subset of roots that obey the specified bounds
signs_cut: signs corresponding to roots_cut
'''
def rootFilter( roots, signs, minVals, maxVals ):
    roots_cut = []
    signs_cut = []
    for r, s in zip( roots, signs ):
        cut = []
        inc = True
        for ri, minr, maxr in zip( r, minVals, maxVals ):
            if ri < minr or ri > maxr:
                inc = False
                break       
        if inc:    
            roots_cut.append( r )
            signs_cut.append( s )
    return roots_cut, signs_cut

'''
A systematic search for stable modules

----------
- Inputs -
----------
update: 1st order ode regulatory functions (takes and returns lists)
t0:     time at which to consider the system
x0:     initial condition to be used in MIOS iterations (should not matter)

minVals, maxVals, tol: defines physical testing bounds and search grid 
                       resolution during root finding
min_ncomb, max_ncomb:  min and max number of positive cycles to combine

plotCharacteristic: determines whether to show characteristic plots during 
                    search

-----------
- Outputs -
-----------
threshVals:   List of lists of valid thresholds
threshStable: -1 the worst case system steady state that generates threshVal is
              stable, 1 otherwise
highNodes:    List of nodes that are above their thresholds in the module
'''
def stableModuleSearch( update, x0, minVals, maxVals, tol, t0 = 0, min_ncomb = 1, max_ncomb = None, plotCharacteristic = False ):
    
    threshVals = []
    threshStable = []
    highNodes = []
    
    # Create regulatory network and find cycles
    G = ODEToDigraph( update, minVals, maxVals, c = x0 )
    cs = cycleSigns( G )
    
    # Extract the cycles that are positive feedback loops
    positiveCycles = []
    for c in cs:
        if c[0] == 1 and len( c[1] ) > 1:
            positiveCycles.append( c[1] )
    
    if max_ncomb is None:
        max_ncomb = len( positiveCycles )
    
    # Consider all combinations of ncomb positive cycles, with ncomb going from
    # min_ncomb to max_ncomb
    for ncomb in range( min_ncomb, max_ncomb + 1 ):
        for pcc in combinations( positiveCycles, ncomb ):
            # Construct a node list and edge list for the cycle combination
            pcl = []
            pc = []
            for c in pcc:
                pc += c 
                pcl += [(c[i],c[i+1]) for i in range(len(c)-1)] + [(c[-1],c[0])]
            pc = list( set( pc ) )
            pcl = list( set( pcl ) )

            # Make sure all subsystem nodes have negative self-regulation
            allNodesDecay = True
            for node in pc:
                if G.has_edge( node, node ):
                    if G[node][node]['weight'] == -1:
                        continue
                allNodesDecay = False
                break            
            if not allNodesDecay:
                continue
            
            # Find the worst case systems and their corresponding regulatory networks
            wcsA, wcsB, hnA, hnB = worstCaseSystem( update, G, pcl, minVals, maxVals )
            GA = ODEToDigraph( wcsA, minVals, maxVals, c = x0 )
            GB = ODEToDigraph( wcsB, minVals, maxVals, c = x0 )
            
            rootsA = []
            signsA = []
            rootsB = []
            signsB = []
            
            # Find the steady states of the worst case systems if the regulatory
            # networks are strongly connected. If they are not, there are two
            # possibilities:
            # 1) The worst case regulation overpowers other regulators, causing
            #    edge deletion in the regulatory network; the worst case is
            #    too bad for a module to stabilize
            # 2) We have combined disjoint cycles; this candidate module is
            #    better considered as smaller modules (lower values of ncomb) 
            
            if nx.is_strongly_connected( GA ):
                rootsA, signsA = MIOSThreshAlgebraic( wcsA, GA, pc, x0, minVals, maxVals, tol, plotCharacteristic = plotCharacteristic )
            if nx.is_strongly_connected( GB ):
                rootsB, signsB = MIOSThreshAlgebraic( wcsB, GB, pc, x0, minVals, maxVals, tol, plotCharacteristic = plotCharacteristic )
                        
            # Filter out steady states that resulted in leaving physical bounds
            # (this should not happen and represents a numerical artifact or
            #  poorly specified bounds)
            rootsCutA, signsCutA = rootFilter( rootsA, signsA, minVals, maxVals )
            rootsCutB, signsCutB = rootFilter( rootsB, signsB, minVals, maxVals )
            
            # Package the filtered roots and signs together with the 
            # specification of which nodes were assumed "high" in constructing
            # the worst case system, then return.
            for r, s in zip( rootsCutA, signsCutA ):
                threshVals.append( r )
                threshStable.append( s )
                highNodes.append( hnA )
                
            for r, s in zip( rootsCutB, signsCutB ):
                threshVals.append( r )
                threshStable.append( s )
                highNodes.append( hnB )
                    
    return threshVals, highNodes, threshStable

'''
Print threshold statements using a sensible formatting.

----------
- Inputs -
----------
thresholds: the steady state of a worst case system, with nan values for 
            uninvolved nodes
highNodes:  a set of nodes assumed to be above threshold; other involved nodes
            are assumed to be below threshold
keys:       a list of strings that give the names of variables by index
            (e.g., keys[0] is the name of the first variable)

-----------
- Outputs -
-----------
None; prints out lines of the form:
name > #
OR
name < #
corresponding to the virtual nodes involved in a stable module.
'''
def printModuleNodes( thresholds, highNodes, keys = None, maxVals = None, roundMultOf = None ):
    if keys is None:
        keys = [str(i) for i in range( len( thresholds ) )]
    if maxVals is None:
        maxVals = np.ones( len( keys ) )
        print("Thresholds as values:")
    else:
        print("Thresholds as fraction of specified maximum values:")
        
        
    for i in range( len( thresholds ) ):
        if thresholds[i] == thresholds[i]:
            thr = thresholds[i]/maxVals[i]
            if not roundMultOf is None:
                thr = round( thr / roundMultOf[i] ) * roundMultOf[i]
            if i in highNodes:
                print( keys[i] + " > " + str( thr ) )
            else:
                print( keys[i] + " < " + str( thr ) )