
import numpy as np
import ODEToDigraph as od

# Define functional forms used in the update function
def H( x, k, n ):
    xkn = ( x / k )**n
    return xkn / ( 1 + xkn )

def Psi( xr, xa, k, n, a0 ):
    return ( a0*k**n + xa**n ) / ( k**n + xa**n + xr**n ) 

# Define the following ordering of variables
en = 0
EN = 1
wi = 2
WI = 3
WE = 4
sp = 5
SP = 6
ci = 7
CA = 8
CR = 9
hh = 10
HH = 11

numVars = 12

# List of variable names coinciding with the above ordering
keys = ["en","EN","wi","WI","WE","sp","SP","ci","CA","CR","hh","HH"]

# Define time constants; these will not affect the stable module analysis,
# and only serve to define the time units.
tau = [0]*numVars
tau[en] = 1
tau[EN] = 1
tau[wi] = 1
tau[WI] = 1
tau[WE] = 1
tau[sp] = 1
tau[SP] = 1
tau[ci] = 1
tau[CA] = 1
tau[CR] = 1
tau[hh] = 1
tau[HH] = 1

# Define kinetic parameters
k_nE = 0.25
n_nE = 2
k_nS = 0.25
n_nS = 2

k_wC = 0.5
n_wC = 4
a_wC = 0.1

k_wI = 0.5
n_wI = 4
a_wI = 0.1

k_in = 0.05
k_out = 0.05

k_sN = 0.25
n_sN = 2

k_cN = 0.25
n_cN = 2

C0 = 0.5
k_CAH = 0.25
n_CAH = 2
k_CRH = 0.25
n_CRH = 2

k_hS = 0.25
n_hS = 2

# Define initial conditions; these do not affect stable module results
u0 = [0]*numVars
u0[en] = 0.5
u0[EN] = 0.5
u0[wi] = 0.5
u0[WI] = 0.5
u0[WE] = 0.5
u0[sp] = 0.5
u0[SP] = 0.5
u0[ci] = 0.5
u0[CA] = 0.5
u0[CR] = 0.5
u0[hh] = 0.5
u0[HH] = 0.5


# There will be two "archetypes"; one for each hedgehog/wingless configuration
# of neighboring cells. These neighboring protein concentrations are constrained 
# by the model to be opposite, so we consider H_Nbr = 1, E_Nbr = 0 (1st 
# archetype) and H_Nbr = 0, E_Nbr = 1 (2nd archetype) 
archList = ["H_Nbr","E_Nbr"]

for archetype in archList:

    # E_nbr and H_nbr must have qualitatively opposite values
    def E_nbr( t ):
        if archetype == "E_Nbr":
            return 1
        else:
            return 0
    def H_nbr(t):
            return 1-E_nbr(t)

    def update( u, t ):
        d = [0]*numVars 
        d[en] = ( H( E_nbr(t), k_nE, n_nE ) * ( 1 - H( u[SP], k_nS, n_nS ) ) - u[en] ) / tau[en]
        d[EN] = ( u[en] - u[EN] ) / tau[EN]
        d[wi] = ( Psi( u[CR], u[CA], k_wC, n_wC, a_wC ) * Psi( 0, u[WI], k_wI, n_wI, a_wI ) - u[wi] ) / tau[wi]
        d[WI] = ( u[wi] - u[WI] ) / tau[WI] + k_in * u[WE] - k_out * u[WI]
        d[WE] = ( -u[WE] ) / tau[WI] + k_out * u[WI] - k_in * u[WE]
        d[sp] = ( 1 - H( u[EN], k_sN, n_sN ) - u[sp] ) / tau[sp]
        d[SP] = ( u[sp] - u[SP] ) / tau[SP]
        d[ci] = ( 1 - H( u[EN], k_cN, n_cN ) - u[ci] ) / tau[ci]
        d[CA] = ( u[ci] - u[CA] ) / tau[CA] - u[CA]/C0 * ( 1 - H( H_nbr(t), k_CAH, n_CAH ) )
        d[CR] = ( -u[CR] ) / tau[CR] + u[CA]/C0 * ( 1 - H( H_nbr(t), k_CAH, n_CAH ) )
        d[hh] = ( 1 - H( u[SP], k_hS, n_hS ) - u[hh] ) / tau[hh]
        d[HH] = ( u[hh] - u[HH] ) / tau[HH]
            
        return d 
    
    maxVals = np.ones( numVars )
    minVals = np.zeros( numVars )
    
    # The max and min values of the cubitus interruptus proteins depend upon
    # the neighboring cell
    if archetype == "H_Nbr":
        minVals[CA] = 0.894
        minVals[CR] = 0.105
        maxVals[CA] = 0.895
        maxVals[CR] = 0.106
    if archetype == "E_Nbr":
        minVals[CA] = 0
        minVals[CR] = 0
        maxVals[CA] = 0.34
        maxVals[CR] = 0.67
        
    tol = [0.01]*numVars
    
    thresh, high, stable = od.stableModuleSearch( update, u0, minVals, maxVals, tol )
    print()
    print()
    print("==========================================")
    print(archetype)
    print("==========================================")
    print()
    for t,s,h in zip( thresh, stable, high ):
        print("-------------------------")
        od.printModuleNodes( t, h, keys = keys )
        print("-------------------------")
    