# An application of our stable module methods to the example in the section
# "Example: Comparison of Phase-Portrait and Stable Module Analyses" of the 
# supporting appendix.

import numpy as np
import ODEToDigraph as od
# Define functional forms used in the update function
def H( x, k=0.25, n=2 ):
    xkn = ( x / k )**n
    return xkn / ( 1 + xkn )


# Define the following ordering of variables
x = 0
y = 1
z = 2

numVars = 3

# List of variable names coinciding with the above ordering
keys = ["x","y","z"]

# Define initial conditions; these do not affect stable module results
u0 = [0]*numVars
u0[x] = 0.5
u0[y] = 0.5
u0[z] = 0.5

def update( uu, t ):
    d = [0]*numVars 
    d[x] = 1-H(uu[y])-uu[x]
    d[y] = (1-H(uu[x]))*H(uu[z])-uu[y]
    d[z] = 0 
    # We do not know the evolution of z; 
    # it will be held fixed in the worst-case system.
    
    return d 

maxVals = np.ones( numVars )
minVals = np.zeros( numVars )
minVals[z] = 0.5125

tol = [0.001]*numVars

thresh, high, stable = od.stableModuleSearch( update, u0, minVals, maxVals, tol )
print()
print()
print("==========================================")
print()
for t,s,h in zip( thresh, stable, high ):
    print("-------------------------")
    od.printModuleNodes( t, h, keys = keys )
    print("-------------------------")
