import matplotlib.pyplot as plt
import scipy as sp
import numpy as np
import networkx as nx
from scipy.integrate import odeint
from math import sin, cos

import ODEToDigraph as od

# Define Hill functions, which are used in the update function
def H( x, k, n ):
    xkn = ( x / k )**n
    
    return xkn / ( 1 + xkn )


# Define the following ordering
TCRb = 0
Fyn  = 1
PAG  = 2
Lck  = 3
TCRp = 4
ZAP  = 5
cCbl = 6

numVars = 7

# List of variable names coinciding with the above ordering
keys = ['TCRb','Fyn','PAG','Lck','TCRp','ZAP','cCbl']

# dx/dt = update( x, t )
def update( x, t ):
    d = [0] * numVars
    
    d[TCRb] = ( 1 - H( x[cCbl], 0.41, 1 ) - x[TCRb] ) / 0.59
    
    d[Fyn]  = ( H( x[Lck], 0.3, 2 ) + H( x[TCRb], 0.27, 6 ) 
               - H( x[Lck], 0.3, 2 ) * H( x[TCRb], 0.27, 6 ) - x[Fyn] ) / 0.46
    
    d[PAG]  = ( H( x[Fyn], 0.69, 1 ) * H( x[TCRb], 0.32, 1 ) 
               + 1 - H( x[TCRb], 0.32, 1 ) - x[PAG] ) / 1.33
    
    d[Lck]  = ( 1 - H( x[PAG], 0.30, 2 ) - x[Lck] ) / 0.15 
    
    d[ZAP]  = ( H( x[Lck], 0.33, 1 ) * H( x[TCRp], 0.29, 2 )
               * ( 1 - H( x[cCbl], 0.3, 1 ) ) - x[ZAP] ) / 2.77
    
    d[cCbl] = ( H( x[ZAP], 0.7, 1 ) - x[cCbl] ) / 0.29
    
    d[TCRp] = ( H( x[TCRb], 0.34, 1 ) * H( x[Lck], 0.34, 1 )
                + H( x[Fyn], 0.13, 1 ) - H( x[TCRb], 0.34, 1 )
                * H( x[Lck], 0.34, 1 ) * H( x[Fyn], 0.13, 1 ) 
                - x[TCRp] ) / 0.03
    
    return d

# Define initial conditions; these do not affect stable module results
x0 = [0]*numVars
x0[TCRb] = 0.5
x0[Fyn]  = 0.5
x0[PAG]  = 0.5
x0[Lck]  = 0.5
x0[ZAP]  = 0.5
x0[cCbl] = 0.5
x0[TCRp] = 0.5

maxVals = np.ones( numVars )
minVals = np.zeros( numVars )
tol = [0.01]*numVars

thresh, high, stable = od.stableModuleSearch( update, x0, minVals, maxVals, tol )
print()
for t,s,h in zip( thresh, stable, high ):
    print("-------------------------")
    od.printModuleNodes( t, h, keys = keys )
    print("-------------------------")

